// Copyright (C) 1998 Logi Ragnarsson

package org.logi.crypto.protocols;
import org.logi.crypto.*;
import org.logi.crypto.keys.*;

import java.math.BigInteger;

/**
 * Diffie-Hellman key exchange server. It expects to talk to a
 * DHKeyExClient object.
 *
 * @see org.logi.crypto.protocols.DHKeyExClient
 *
 * @author <a href="http://www.logi.org/~logir/">Logi Ragnarsson</a>
 * (<a href="mailto:logir@logi.org">logir@logi.org</a>)
 */
public class DHKeyExServer extends DHKeyEx implements InterKeyExServer {
    
    /**
     * Create a new DHKeyExServer object. It uses
     * the private DH key from <code>pri</code>.
     *
     * @exception KeyException if the key is not private. */
    public DHKeyExServer(DHKey pri, String keyType) throws KeyException {
        super(pri,keyType);
    }
    
    /**
     * Create a new DHKeyExClient object. It uses
     * an <code>n</code> bit modulus and the named key type.
     * <p>
     * There are pre-computed public modulus and gnerator pairs for
     * these values of <code>n</code>: 256, 512, 1024, 2048. Using
     * one of these values saves you from a rather long wait.
     */
    public DHKeyExServer(int n, String keyType){
        super(n,keyType);
    }
    
    /**
     * Get the next message in the protocol.
     * <p>
     * <code>received</code> is the last message received form the client
     * and has not yet been sent to the client.
     * <p>
     * The returned value is the next message to send to the client or null
     * if no more messages need to be sent and the protocol is terminated.
     *
     * @exception CryptoProtocolException if a problem arises with the protocol.
     */
    public byte[] message(byte[] received) throws CryptoProtocolException {
	try{

	    if(received!=null){
		hisPublic = new BigInteger(1,received);
		byte[] k= hisPublic.modPow(myPrivate,m).toByteArray();
		try{
		    sessionKey=makeSessionKey(keyType,k);
		    keyDecided=true;
		} catch (InvalidCDSException e) {
		    throw new CryptoProtocolException(e.getMessage());
		}
	    }
	    byte[] r = myPublic.toByteArray();
	    if(r.length==modSize/8)
		return r;
	    byte[] s = new byte[modSize/8];
	    int l=Math.min(r.length, modSize/8);
	    System.arraycopy(r, r.length-l, s,s.length-l, l);
	    return s;

	} catch (ArrayIndexOutOfBoundsException e) {
	    throw new CryptoProtocolException("Malformed message.");
	}
    }
    
}
